from dataclasses import dataclass
import os
import shutil
import tempfile
import logging

from video_utils import Video

from .ffmpeg_converter import FFmpegConverter
from .settings import AudioSettings, VideoSettings

log = logging.getLogger(__name__)


@dataclass
class VideoProcessor:
    video: Video
    video_settings: VideoSettings
    audio_settings: AudioSettings
    container: str
    dry_run: bool = False
    in_place: bool = False
    extra_ffmpeg_input_args: str = ""
    extra_ffmpeg_output_args: str = ""
    temp_directory: str = None

    # def __post__init__(self):
    #     self.temp_file = self._create_temp_file()

    def _create_temp_file(self):
        return tempfile.mkstemp(dir=self.temp_directory, suffix=f".{self.container}")[1]

    def __str__(self):
        return f"Video: {self.video.full_path}, format: {self.video.codec.pretty_name}, quality: {self.video.quality}"

    def process(self):
        self.temp_file = self._create_temp_file()
        converter = FFmpegConverter(
            source_file_path=self.video.full_path,
            destination_file_path=self.temp_file,
            extra_ffmpeg_input_args=self.extra_ffmpeg_input_args,
            extra_ffmpeg_output_args=self.extra_ffmpeg_output_args,
            video_settings=self.video_settings,
            audio_settings=self.audio_settings,
            dry_run=self.dry_run,
        )
        converter.process()
        self._move_output_video()

    def already_processed(self):
        renamed_path = self.renamed_path()
        if os.path.exists(renamed_path):
            print(
                f"File {self.video.name} appears to have already been converted to {renamed_path} exists. Skipping")
            return True

        split_filename = os.path.splitext(self.video.name)
        codec_name = self.video_settings.codec.pretty_name
        if split_filename[0].endswith(codec_name):
            print(
                f"File {self.video.name} already matches the renaming format. Skipping")
            return True

        return False

    def renamed_path(self):
        split_filename = os.path.splitext(self.video.full_path)
        codec_name = self.video_settings.codec.pretty_name
        return f"{split_filename[0]} - {codec_name}.{self.container}"

    def in_place_file_path(self):
        split_filename = os.path.splitext(self.video.full_path)
        return f"{split_filename[0]}.{self.container}"

    def _move_output_video(self):
        if not self.dry_run:
            shutil.move(self.temp_file, self.renamed_path())
        if self.in_place:
            if self.dry_run:
                print(f"DRY-RUN: Would replace original file {self.video.full_path}")
                return
            print(f"Replacing original file {self.video.full_path}")
            os.remove(self.video.full_path)
            shutil.move(self.renamed_path(), self.in_place_file_path())



import click
import logging

from video_utils import Codec

from .processor import Processor, AudioSettings, VideoSettings

# TODO: How do I set show_defaults on all commands?
CONTEXT_SETTINGS = dict(help_option_names=['--help', '-h'])


def set_log_level(verbose):
    log_level = logging.WARNING
    if verbose:
        log_level = logging.INFO
    logging.basicConfig(level=log_level)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("directory",
                type=click.Path(exists=True),
                required=True)
@click.option("--in-place", "-i", is_flag=True,
              help="Replace the original files instead of appending the new codec name")
@click.option(
    "--force", "-f", is_flag=True,
    help="Force conversion even if the format of the file already matches the desired format")
@click.option("--video-codec", default="HEVC", show_default=True,
              help="A target video codec. Supported codecs: HEVC, HEVC_NVENC, AVC")
@click.option("--quality", "-q", type=int, default=22, show_default=True, help="The quantizer quality level to use")
@click.option("--preset", "-p", default="faster", show_default=True, help="FFmpeg preset to use")
@click.option("--width", "-w", type=int, help="Specify a new width to enable resizing of the video")
@click.option("--extra-input-args", default="",
              help="Specify any extra arguments you would like to pass to FFMpeg input here")
@click.option("--extra-output-args", default="",
              help="Specify any extra arguments you would like to pass to FFMpeg output here")
@click.option("--audio-codec", default="AAC", show_default=True, help="A target audio codec")
@click.option("--audio-channels", default=2, show_default=True, type=int,
              help="The number of channels to mux sound in to")
@click.option("--audio-bitrate", default=160, show_default=True, type=int,
              help="The bitrate to use for the audio codec")
@click.option("--temp-dir", help="Specify a temporary directory to use during conversions instead of the system default")
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose log output")
@click.option("--container", default="mkv", show_default=True,
              help="Specify a video container to convert the videos in to")
@click.option("--dry-run", is_flag=True, help="Do not make actual changes")
def main(directory, force, in_place, video_codec,  quality, preset,
         width, extra_input_args, extra_output_args, audio_codec, audio_channels,
         audio_bitrate, temp_dir, verbose, container, dry_run):

    set_log_level(verbose)

    video_settings = VideoSettings(
        codec=Codec(video_codec),
        quality=quality,
        preset=preset,
        width=width,
    )
    audio_settings = AudioSettings(
        codec=Codec(audio_codec),
        channels=audio_channels,
        bitrate=audio_bitrate,
    )

    if video_codec.Codec.ffmpeg_name is None:
        raise Exception("Invalid video codec specified")

    if audio_codec.Codec.ffmpeg_name is None:
        raise Exception("Invalid audio codec specified")

    Processor(
        directory=directory,
        force=force,
        video_settings=video_settings,
        audio_settings=audio_settings,
        in_place=in_place,
        extra_ffmpeg_input_args=extra_input_args,
        extra_ffmpeg_output_args=extra_output_args,
        temp_directory=temp_dir,
        container=container,
        dry_run=dry_run,
    ).start()



from dataclasses import dataclass
import logging
import os

import ffmpy

from .settings import AudioSettings, VideoSettings


log = logging.getLogger(__name__)


@dataclass
class FFmpegConverter:
    source_file_path: str
    destination_file_path: str
    extra_ffmpeg_input_args: str
    extra_ffmpeg_output_args: str
    dry_run: bool
    video_settings: VideoSettings
    audio_settings: AudioSettings

    def __post__init__(self):
        self._validate_destination()

    def process(self):
        input_settings = self._generate_ffmpeg_settings("input")
        output_settings = self._generate_ffmpeg_settings("output")

        ff = ffmpy.FFmpeg(
            inputs={self.source_file_path: input_settings},
            outputs={self.destination_file_path: output_settings})
        if self.dry_run:
            print(f"DRY-RUN: Would start conversion. Command: '{ff.cmd}'")
        else:
            log.info(f"Starting conversion. Command: '{ff.cmd}'")
            ff.run()

    def _generate_ffmpeg_settings(self, mode):
        if mode == "input":
            return self.extra_ffmpeg_input_args

        output_settings = "" + \
            "-y" + \
            " -threads 0" + \
            str(self.video_settings) + \
            str(self.audio_settings) + \
            " " + self.extra_ffmpeg_output_args
        return output_settings

    def _validate_destination(self):
        if os.path.exists(self.destination_file_path):
            if os.path.isfile(self.destination_file_path):
                if os.access(self.destination_file_path, os.W_OK):
                    return True
                raise IOError(
                    f"Destination file {self.destination_file_path} exists but is not writable")
            if os.path.isdir(self.destination_file_path):
                raise IOError(
                    f"Destination file {self.destination_file_path} is a directory, and not a file")
        else:
            try:
                with open(self.destination_file_path, 'w') as f:
                    f.writelines("test")
            finally:
                os.remove(self.destination_file_path)
            return True
